package de.hftstuttgart.unifiedticketing.systems.github;

import com.fasterxml.jackson.databind.ObjectMapper;
import de.hftstuttgart.unifiedticketing.core.Logging;
import de.hftstuttgart.unifiedticketing.exceptions.AssertionException;
import de.hftstuttgart.unifiedticketing.exceptions.DeserializationException;
import de.hftstuttgart.unifiedticketing.exceptions.HttpResponseException;
import okhttp3.*;
import okio.Buffer;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mockito;

import java.io.IOException;
import java.util.*;
import java.util.logging.Level;
import java.util.stream.Collectors;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

public class GithubTicketTest
{
    public GithubTicket instance;
    public Call call;
    public ArgumentCaptor<Request> requestCaptor;
    public Response.Builder responseBuilder;

    @BeforeAll
    public static void initBeforeAll()
    {
        Logging.setLevel(Level.ALL);
    }

    @BeforeEach
    public void initBeforeEach()
    {
        instance =
            spy(new GithubTicket(
                spy(new GithubTicketSystem("someHeader", "https://example.org"))));

        OkHttpClient client = mock(OkHttpClient.class);
        call = mock(Call.class);
        requestCaptor = ArgumentCaptor.forClass(Request.class);
        responseBuilder = new Response.Builder()
            .request(new Request.Builder().url("http://test.some.tld/").build())
            .protocol(Protocol.HTTP_1_1)
            .message("some message");

        doReturn(client).when(instance).getHttpClient();
        doReturn(call).when(client).newCall(requestCaptor.capture());
    }

    @Test
    public void testFromTicketResponse()
    {
        GithubTicketResponse res = new GithubTicketResponse();
        GithubTicket ticket = GithubTicket.fromTicketResponse(instance.getParent(), res);

        assertAll(
            () -> assertSame(instance.getParent(), ticket.getParent()),
            () -> assertNull(ticket.getTitle()),
            () -> assertNull(ticket.getDescription()),
            () -> assertEquals("0", ticket.getId()),
            () -> assertNotNull(ticket.getLabels()),
            () -> assertTrue(ticket.isOpen()),
            () -> assertNull(ticket.getTitle()),
            () -> assertNotNull(ticket.getAssignees())
        );
    }

    @Test
    public void testAddAssigneeByString()
    {
        String username = "someUsername";
        instance.addAssignee(username);
        assertEquals(1, instance.getAssignees().size());
        assertEquals(username, instance.getAssignees().stream().findFirst().get().username);
    }

    @Test
    public void testRemoveAssigneeByString()
    {
        // first add assignee and check it was added properly
        testAddAssigneeByString();

        String username = instance.getAssignees().stream().findFirst().get().username;
        instance.removeAssignee(username);
        assertEquals(0, instance.getAssignees().size());
    }

    @Test
    public void testSaveNoHttpClientWithoutChanges()
    {
        GithubTicket actual = instance.save();
        verify(instance, never()).getHttpClient();
        assertSame(instance, actual);
    }

    @Test
    public void testSaveServerError() throws IOException
    {
        doReturn(responseBuilder.code(500).build()).when(call).execute();
        instance.setTitle("we have a new title");
        assertThrows(HttpResponseException.class, () -> instance.save());
    }

    @Test
    public void testSaveClientError() throws IOException
    {
        doReturn(responseBuilder.code(400).build()).when(call).execute();
        instance.setTitle("we have another new title");
        assertThrows(HttpResponseException.class, () -> instance.save());
    }

    @Test
    public void testSaveNullBody() throws IOException
    {
        doReturn(responseBuilder.code(200).build()).when(call).execute();
        instance.setTitle("we have no body this time");
        assertThrows(HttpResponseException.class, () -> instance.save());
    }

    @Test
    public void testSaveNoJsonBody() throws IOException
    {
        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create("somestrangething", MediaType.get("application/json")))
                .build())
            .when(call).execute();
        instance.setTitle("this body is no json");
        assertThrows(DeserializationException.class, () -> instance.save());
    }

    @Test
    public void testSaveSuccessfulUpdate() throws IOException
    {
        GithubTicketResponse ticketResponse = new GithubTicketResponse();
        ticketResponse.title = "title of ticket";
        ticketResponse.body = "description";
        ticketResponse.number = 5;
        ticketResponse.labels = Arrays.stream(new String[]{"bug", "unifiedticketing"})
            .map(l -> {
                GithubTicketResponse.Label label = new GithubTicketResponse.Label();
                label.name = l;
                return label;
            })
            .collect(Collectors.toSet());
        GithubTicketResponse.Assignee assignee = new GithubTicketResponse.Assignee();
        assignee.id = 234;
        assignee.login = "username";
        ticketResponse.assignees = new HashSet<>(Collections.singleton(assignee));

        GithubTicket expected = GithubTicket.fromTicketResponse(instance.getParent(), ticketResponse);
        String responseJson = new ObjectMapper().writeValueAsString(ticketResponse);

        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create(responseJson.getBytes(), MediaType.get("application/json")))
                .build())
            .when(call).execute();
        GithubTicket actual = instance.close().save();
        assertEquals(expected, actual);
        assertTrue(expected.deepEquals(actual));
    }

    @Test
    public void testSaveRequestJson() throws IOException
    {
        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create("{}".getBytes(), MediaType.get("application/json")))
                .build())
            .when(call).execute();

        instance.setTitle("title of ticket")
            .setDescription("description")
            .setLabels(new HashSet<>(Arrays.asList("bug", "unifiedticketing")))
            .addAssignee("someUsername")
            .save();

        Buffer buffer = new Buffer();
        ObjectMapper mapper = new ObjectMapper();
        requestCaptor.getValue().body().writeTo(buffer);
        String expectedJson = "{\"assignees\":[\"someUsername\"],\"body\":\"description\"," +
            "\"title\":\"title of ticket\",\"labels\":[\"bug\",\"unifiedticketing\"]}";
        assertEquals(mapper.readTree(expectedJson), mapper.readTree(buffer.readUtf8()));
    }
}
